@@ -1,4 +1,4 @@
-# Copyright (C) 2013 10gen Inc.
+# Copyright (C) 2009-2013 MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -20,14 +20,14 @@ class CursorTest < Test::Unit::TestCase
   include Mongo::Constants
 
   @@connection = standard_connection
-  @@db   = @@connection.db(MONGO_TEST_DB)
+  @@db   = @@connection.db(TEST_DB)
   @@coll = @@db.collection('test')
   @@version = @@connection.server_version
 
   def setup
     @@coll.remove
     @@coll.insert('a' => 1)     # collection not created until it's used
-    @@coll_full_name = "#{MONGO_TEST_DB}.test"
+    @@coll_full_name = "#{TEST_DB}.test"
   end
 
   def test_alive
@@ -91,6 +91,41 @@ class CursorTest < Test::Unit::TestCase
     end
   end
 
+  def test_compile_regex_get_more
+    return unless defined?(BSON::BSON_RUBY) && BSON::BSON_CODER == BSON::BSON_RUBY
+    @@coll.remove
+    n_docs = 3
+    n_docs.times { |n| @@coll.insert({ 'n' => /.*/ }) }
+    cursor = @@coll.find({}, :batch_size => (n_docs-1), :compile_regex => false)
+    cursor.expects(:send_get_more)
+    cursor.to_a.each do |doc|
+      assert_kind_of BSON::Regex, doc['n']
+    end
+  end
+
+  def test_max_time_ms_error
+    cursor = @@coll.find
+    cursor.stubs(:send_initial_query).returns(true)
+
+    cursor.instance_variable_set(:@cache, [{
+      '$err' => 'operation exceeded time limit',
+      'code' => 50
+    }])
+
+    assert_raise ExecutionTimeout do
+      cursor.to_a
+    end
+  end
+
+  def test_max_time_ms
+    with_forced_timeout(@@connection) do
+      assert_raise ExecutionTimeout do
+        cursor = @@coll.find.max_time_ms(100)
+        cursor.to_a
+      end
+    end
+  end
+
   def test_exhaust_after_limit_error
     c = Cursor.new(@@coll, :limit => 17)
     assert_raise MongoArgumentError do
