@@ -1,4 +1,4 @@
-# Copyright (C) 2013 10gen Inc.
+# Copyright (C) 2009-2013 MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -19,7 +19,7 @@ class DBAPITest < Test::Unit::TestCase
   include BSON
 
   @@client = standard_connection
-  @@db   = @@client.db(MONGO_TEST_DB)
+  @@db   = @@client.db(TEST_DB)
   @@coll = @@db.collection('test')
   @@version = @@client.server_version
 
@@ -27,7 +27,7 @@ class DBAPITest < Test::Unit::TestCase
     @@coll.remove
     @r1 = {'a' => 1}
     @@coll.insert(@r1) # collection not created until it's used
-    @@coll_full_name = "#{MONGO_TEST_DB}.test"
+    @@coll_full_name = "#{TEST_DB}.test"
   end
 
   def teardown
@@ -325,14 +325,14 @@ class DBAPITest < Test::Unit::TestCase
     @@db.strict = true
 
     begin
-      coll = @@db.create_collection('foobar', :capped => true, :size => 1024)
+      coll = @@db.create_collection('foobar', :capped => true, :size => 4096)
       options = coll.options
-      assert_equal 'foobar', options['create']
+      assert_equal 'foobar', options['create'] if @@client.server_version < '2.5.5'
       assert_equal true, options['capped']
-      assert_equal 1024, options['size']
+      assert_equal 4096, options['size']
     rescue => ex
       @@db.drop_collection('foobar')
-      fail "did not expect exception \"#{ex}\""
+      fail "did not expect exception \"#{ex.inspect}\""
     ensure
       @@db.strict = false
     end
